# Created by: Ausar686
# https://github.com/Ausar686

from .qagpt import QAGPT


class AntiListActor(QAGPT):

	_prompt = """
		###INSTRUCTIONS###
		Inspect the text below.
		If there is a numerated list in text write only the first element of the list, explained in detail, according to context.
		Keep everything else (except of numerated list) exactly the same, as the original.
		Keep the same language, as the original.
		Keep the same style, as the original.
		Example:
			Input: 
				There are multiple sports, that can suit you. Here are several examples:
					1. Football
					2. Hockey
					3. Basketball
					4. Volleyball
					5. Tennis
				It's crucial to choose a sport, according to personal preferences. What do you think about it?
			Output:
				There are multiple sports, that can suit you. For example, football is a decent solution because of your outstanding stamina. What do you think about it?

		###TEXT###
	"""

	def run(self, text: str) -> str:
		request = f"{self._prompt}{text}"
		result = self.get_str(request)
		enum_str = "\n\n1."
		index = result.find(enum_str)
		if index >= 0:
			result = result[:index] + result[index+len(enum_str)+1:]
		return result


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union, Any
import json
from collections import defaultdict


class UserData:
    _attributes = ["first_name", "last_name", "username", "_fav_topics", "_fav_diary_topics", "_fav_actions_topics"]
    
    def __init__(self, data: Union[str, dict]=None):
        if data is None:
            self.from_empty()
            return
        elif isinstance(data, str):
            self.from_json(data)
            return
        elif isinstance(data, dict):
            self.from_dict(data)
            return
        else:
            raise TypeError(f"Can't initialize UserData object from type {type(data)}"
                           "Supported types are: str, dict.")
    
    def from_json(self, path: str):
        with open(path, "r", encoding="utf-8") as json_file:
            dct = json.load(json_file)
        self.from_dict(dct)
        return
    
    def from_empty(self) -> None:
        for attr in self._attributes:
            if attr[0] == "_":
                self.__setattr__(attr, defaultdict(float))
            else:
                self.__setattr__(attr, None)
        return
    
    def from_dict(self, dct: dict) -> None:
        for attr in self._attributes:
            if attr in dct:
                if attr[0] == "_":
                    self.__setattr__(attr, defaultdict(float, dct[attr]))
                else:
                    self.__setattr__(attr, dct[attr])
            else:
                if attr[0] == "_":
                    self.__setattr__(attr, defaultdict(float))
                else:
                    self.__setattr__(attr, None)
        return
    
    @property
    def dict(self) -> dict:
        return {attr: self.__getattribute__(attr) for attr in self._attributes}
    
    def to_json(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as json_file:
            json.dump(self.dict, json_file)
        return
    
    def __getitem__(self, item: str) -> Any:
        # An alias for __getattribute__.
        return self.__getattribute__(item)
    
    def load(self, path: str) -> None:
        # An alias for 'from_json' for standartized interface
        self.from_json(path)
        return
    
    def to_disk(self, path: str) -> None:
        # An alias for 'to_json' for standartized interface
        self.to_json(path)
        return


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union, Any

from pandas.core.frame import DataFrame

from .base_storage import BaseStorage


class Actions(BaseStorage):
    
    def __init__(self, df: Union[DataFrame,str]=None):
        self._columns = ["media", "author", "title", "topic", "duration", "text"]
        self._indexed_columns = ["media", "author", "title", "topic"]
        self._multivalue_columns = ["topic"]
        super().__init__(df)
        return
    
    def get(self, 
            *, 
            by: str, 
            value: Any=None, 
            search_type: str="union", 
            date_search: str="precise") -> DataFrame:
        ret, data = self._base_get(
            by=by,
            value=value,
            search_type=search_type,
            date_search=date_search)
        if ret:
            return data
        if by == "media":
            return self._eq_search(by, value)
        elif by == "author":
            return self._eq_search(by, value)
        elif by == "title":
            return self._eq_search(by, value)
        elif by == "topic":
            return self._substr_search(by, value)
        else:
            raise ValueError("Invalid search argument.")


# Created by: Ausar686
# https://github.com/Ausar686

from bs4 import BeautifulSoup
import requests
import urllib


class GoogleSearcher:
    """
    RAI class, that implements Web browsing 
    via Google Search Engine.
    """
    
    def __init__(self):
        """
        Initializes instance with attributes to store data
        for debugging.
        """
        self.clear()
        self.google = "https://google.com/search"
        self.prefix = "/url?q="
        self.suffix = "&sa="
        
    def get_response(self, query: str) -> requests.Response:
        """
        Gets response from GET requestto Google Search Engine.
        """
        # Set query parameter
        params = {"q": query}
        # Get response from Google Search
        resp = requests.get(self.google, params=params)
        # Save response to 'resp' attribute
        self.resp = resp
        return resp
    
    def fetch_urls(self, resp: requests.Response) -> list:
        """
        Fetches urls from response, using BeautifulSoup from bs4 library.
        """
        # Get response content as text (not binary)
        content = resp.text
        # Parse the content using BeautifulSoup
        soup = BeautifulSoup(content, 'html.parser')
        # Extracting URLs based on the provided structure
        search_result_anchors = soup.find_all('a')
        # Getting the href attribute from these anchors
        page_urls = [a['href'] for a in search_result_anchors]
        # Getting urls, that start with "/url?q="
        result_urls = [url.split(self.prefix)[-1] for url in page_urls if url.startswith(self.prefix)]
        return result_urls
    
    @staticmethod
    def is_google_url(url: str) -> bool:
        """
        Checks, whether the URL is not a result of search,
        and is a Google URL instead 
        """
        return (
            url.find("support.google.com") >= 0 or
            url.find("accounts.google.com") >= 0)
    
    def postprocess_urls(self, urls: list) -> list:
        """
        Performs URLs postprocessing, so that to return clear urls,
        which are ready-to-use by end user.
        """
        # Strip urls
        stripped_urls = [url[:url.find(self.suffix)] if url.find(self.suffix) != -1 else url for url in urls]
        # Unquote urls
        unquoted_urls = [urllib.parse.unquote(url) for url in stripped_urls]
        # Remove all results with support.google.com and accounts.google.com
        cleared_urls = [url for url in unquoted_urls if not self.is_google_url(url)]
        self.urls = cleared_urls
        return cleared_urls
    
    def clear(self) -> None:
        """
        Clears previous results (self.resp and self.urls)
        """
        self.resp = None
        self.urls = None
        return
    
    def search(self, query: str) -> list:
        """
        Runs a full search pipeline: 
        1. 'get_response'
        2. 'fetch_urls'
        3. 'postprocess_urls'
        """
        r = self.get_response(query)
        if r.status_code == 200:
            urls = self.fetch_urls(r)
            urls = self.postprocess_urls(urls)
            return urls
        else:
            print(f"[ERROR]: URL: {r.url} STATUS: {r.status_code}")
            return None
    
    def run(self, query: str) -> list:
        """
        An alias for 'search' method for compatibility with other actors
        """
        return self.search(query)


# Created by: Ausar686
# https://github.com/Ausar686

from .actors import *
from .chat import *
from .containers import *
from .heuristics import *
from .profile import *
from .storages import *
from .utils import *


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Union
import os
import re
import time

from .base_actor import BaseActor
from ..containers.rdict import RDict
from ..utils import request_openai, to_rdict


class QAGPT(BaseActor):
    """
    Wrapping class around OpenAI API to simplify development and automatic usage.
    """
    # TODO: Test and improve wrapping prompts.
    # TODO: Implement streaming output.
    # TODO: Implement processing of different 'n' values.
    # TODO: Check implementation of methods, so that to use static/class methods where needed. 
    def __init__(
        self, 
        model: str='gpt-3.5-turbo', 
        key: str=None,
        temperature: float=0,
        stream: bool=False,
        n: int=1,
        ):
        """
        Initializes QAGPT(Qusetion-Answer GPT) object instance.
        Args:
            model (str): Type of model for API to use. Default: 'gpt-3.5-turbo'
            key (str): OpenAI API key. By default it's supposed, that you store it inside of 'OPENAI_API_KEY' environment variable. Default: None.
            temperature (float): Randomness of the results. The higher the value is, the more creative answers you will get.
                                 Lower values, yet, provide conservative answers. Should be in interval [0,2]. Default: 0.
            stream (bool): Whether to provide a omplete answer at once (False), or token-by-token (True). Default: False.
            n (int): Number of answers to generate. NOTE: Each answer consumes tokens. Default: 1. 
        """
        super().__init__(model)
        if stream:
            raise NotImplementedError("Streaming option coming soon...")
        if n > 1:
            raise NotImplementedError("Multiple output coming soon...")
        # Init OpenAI parameters
        self.openai = RDict()
        self.set_api_key(key)
        self.openai.model = self.model
        self.openai.temperature = temperature
        self.openai.stream = stream
        self.openai.n = n
        self._info = f"Instance of wrapping class around OpenAI API to simplify development and automatic usage.\nModel version: {self.model}"
        self._sleep_time = 2
        self._supported_types = ['int', 'float', 'list', 'dict', 'str']
        self._n_supported_types = len(self._supported_types)
        self._supported_roles = ['user', 'system', 'assistant']
        self._wrap_int_string = "Write only the integer as the answer."
        self._wrap_float_string = "Write only the float as the answer."
        self._wrap_prob_string = "Write only the float from interval [0, 1] as the answer. The closer the result is to 0, the less the probability is. The closer the result is to 1, the greater the probability is."
        self._wrap_dict_string = "Write the answer in Python dictionary notation. Write only the pairs key:value from the dictionary, separated by newline symbol."
        self._wrap_list_string = "Write only the list. Each element of the list should be written on a separate line."
        
    def set_api_key(self, key: str) -> None:
        """
        Method, that sets openai.api_key value.
        The API key is set only if it was None before calling the method. 
        Args:
            key (str): Value, that should be set as openai.api_key. If key is None gets an environmental variable 'OPENAI_API_KEY'
        Returns:
            None
        """
        if key is None:
            key = os.getenv("OPENAI_API_KEY")
        self.openai.api_key = key
        return

    def set_model(self, model: str) -> None:
        """
        Override for a base class method in order to remain compatible 
        with new format of RDict 'openai' as an attribute
        """
        super().set_model(model)
        if hasattr(self, "openai"):
            self.openai.model = model
        return
        
    def form_message(self, request: str, role: str="user") -> dict:
        """
        Takes text request and a given role in GPT-API and transforms it into a dictionary.
        Args:
            request (str): Text request to OpenAI API.
            role (str): String, which represents the role. Should be in ['user', 'system', 'assistant'].
                For further info about roles check GPT API documentation or community guides
                like this: https://vc.ru/u/1389654-machine-learning/625306-prostoe-rukovodstvo-po-chatgpt-api-s-ispolzovaniem-python
        Returns:
            message (dict): A dictionary for further processing with GPT-API.
        """
        assert role in self._supported_roles
        return {"role":role, "content":request}
    
    def form_messages(self, request: str, role: str="user") -> list:
        """
        Forms and returns a list from message, made with 'form_message' method.
        Args:
            request (str): Text request to OpenAI API.
            role (str): String, which represents the role. Should be in ['user', 'system', 'assistant'].
                For further info about roles check OpenAI API documentation: https://platform.openai.com/docs/api-reference/chat/create
        Returns:
            messages (list): List with dictionary (messages) made with 'form_message' method.
        """
        return [self.form_message(request, role)]
    
    def ask(self, messages: list) -> str:
        """
        Sends request to OpenAI server and returns a answer in string format.
        Args:
            messages (list): List of messages (dicts). By default, it's supposed, that you make messages using 'form_messages' method.
        Reurns:
            answer (str): Model response in string format.
        """
        json_data = request_openai(messages=messages, **self.openai)
        completion = to_rdict(json_data)
        answer = completion.choices[0].message.content
        return answer
    
    def get_answer(self, messages: list) -> str:
        """
        Wrapper around 'ask' method for case of running out of TPM (tokens per minute) or RPM (requests per minute).
        It is strongly recommended to use this method and not 'ask' one.
        Args:
            messages (list): List of messages (dicts). By default, it's supposed, that you make messages using 'form_messages' method.
        Reurns:
            answer (str): Model response in string format.
        """
        while True:
            try:
                answer = self.ask(messages)
                return answer
            except Exception:
                print(f"[INFO]: TPM/RPM limit exceeded. Waiting {self._sleep_time} seconds to continue requesting.")
                time.sleep(self._sleep_time)

    # Here comes a bunch of prompt wrappings for obtaining an answer of specific type.
    # NOTE: It's not supposed, that you will use this methods manually.
    # Use 'get_' methods instead to obtain a result of a specific type.
    
    def wrap_int(self, request: str) -> str:
        return request + self._wrap_int_string
    
    def wrap_float(self, request: str) -> str:
        return request + self._wrap_float_string
    
    def wrap_prob(self, request: str) -> str:
        return request + self._wrap_prob_string
    
    def wrap_dict(self, request: str) -> str:
        return request + self._wrap_dict_string
    
    def wrap_list(self, request: str) -> str:
        return request + self._wrap_list_string
    
    def wrap_type(self, request: str) -> str:
        """
        EXPERIMENTAL: Wraps request, so that OpenAI API will return the name of type of the answer, that user wants to obtain. May work inaccurately.
        """
        return f"Imagine you get a request: '{request}'. What type of answer does the user expect:{', '.join(self._supported_types)}? Write the dictionary in Python notation as the answer. The dictionary contains {self._n_supported_types} keys (all of them are strings): {', '.join(self._supported_types)}; and values are the probabilities of these types."
    
    def scrub_elem(self, elem: str) -> str:
        """
        Scrubs an element from the list, returned by API as an answer.
        This function strips the elem and then removes it's index from the beginning.
        Example: " 123. Football " -> "Football"
        Args:
            elem (str): Element of the list, obatined by parsing an answer, returned by OpenAI API.
        Returns:
            scrubbed_elm (str): Stripped 'elem' without index in the beginning. 
        """
        return re.sub(r'(?:\d*\.|-)\s*(.*)', r'\1',  elem.strip())
    
    def get_int(self, request: str) -> int:
        """
        Return an API response for the request as an integer.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (int): OpenAI API response as an integer.
        """
        while True:
            try:
                return int(self.get_answer(self.form_messages(self.wrap_int(request))))
            except Exception:
                pass
    
    def get_float(self, request: str) -> float:
        """
        Return an API response for the request as a float.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (float): OpenAI API response as a float.
        """
        while True:
            try:
                return float(self.get_answer(self.form_messages(self.wrap_float(request))))
            except Exception:
                pass
    
    def get_prob(self, request: str) -> float:
        """
        Return an API response for the request as a probability (float from 0.0 to 1.0).
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (float): OpenAI API response as a probability.
        """
        while True:
            try:
                answer = self.get_float(request)
                assert 0 <= answer <= 1
                return answer
            except Exception:
                pass
    
    def get_list(self, request: str) -> list:
        """
        Return an API response for the request as a list.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (list): OpenAI API response as a list.
        """
        while True:
            try:
                return list(map(self.scrub_elem, self.get_answer(self.form_messages(self.wrap_list(request))).split('\n')))
            except Exception:
                pass
    
    def get_dict(self, request: str) -> dict:
        """
        Return an API response for the request as an integer.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (dict): OpenAI API response as a dict.
        """
        # [TODO]: Add functionallity to process answers with "".
        while True:
            try:
                res_dict = {}
                answer = self.get_answer(self.form_messages(self.wrap_dict(request)))
                # Parse answer string with regexp, so that to get a list of strings with key/value pairs for the dict.
                key_val_array = re.findall(r"'[^']*':[^,]*(?:,|$)", re.sub(r"[\{\}]", "", answer))
                for elem in key_val_array:
                    # Parse each string in list to obtain key/value independently.
                    value = re.sub(r'[,\n]$', '', re.sub(r"'[^']*':\s*", '', elem))
                    key = re.findall(r"'[^']*':", elem)[0][1:-2]
                    res_dict[key] = value
                return res_dict
            except Exception:
                pass
    
    def get_str(self, request: str) -> str:
        """
        Return an API response for the request as a string.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (str): OpenAI API response as a string.
        """
        return self.get_answer(self.form_messages(request))
    
    def get_type(self, request: str) -> str:
        """
        [EXPERIMENTAL]: Returns the name of type of the answer, that user wants to obtain. May work inaccurately.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            max_type (str): Predicted name of type of the answer, that user wants to obtain.
        """
        prob_dict = self.get_dict(self.wrap_type(request))
        prob_dict = {key:float(value) for key, value in prob_dict.items()}
        max_type = [type_ for type_ in prob_dict if prob_dict[type_] == max(prob_dict.values())][0]
        return max_type
    
    def __call__(self, request: str) -> Union[int, float, list, dict, str]:
        """
        [EXPERIMENTAL]: Returns an OpenAI API response to request in type, that user wants to obtain. May work inaccurately.
        Args:
            request (str): Text request to OpenAI API.
        Returns:
            answer (Union[int, float, list, dict, str]): OpenAI API answer in a most desired format.
        """ 
        return getattr(self, f'get_{self.get_type(request)}')(request)

    def run(self, request: str) -> str:
        """
        An alias for 'get_str' method for usage as an actor in ChatBot
        """
        return self.get_str(request)

    def __repr__(self):
        """
        Returns an information about the instance of class.
        """
        return self._info

    def __str__(self):
        """
        An alias for '__repr__' method.
        """
        return self.__repr__()


from .pmc_models import ModelPMC, ModelPubMedLine
import orjson
import requests
from requests.utils import quote
from pydantic import parse_obj_as

class Singleton(type):
  _instances = {}
  def __call__(cls, *args, **kwargs):
      if cls not in cls._instances:
          instance = super().__call__(*args, **kwargs)
          cls._instances[cls] = instance
      return cls._instances[cls]
  
class PMCRequester(metaclass=Singleton):
    """
    The `ModelPMC` instance is constructed by invoking its constructor with arguments that are parsed from the `source` dictionary 
    using the `parse_obj_as` function. The `parse_obj_as` function takes two arguments: a type hint for the expected data type, 
    and the actual data. It returns the parsed object if parsing succeeds, or raises an error if parsing fails. 
    The `ModelPMC` class is defined somewhere else in the code and likely contains data attributes 
    that correspond to the keys in the `source` dictionary.
    """
    def __init__(self, ip_address: str, port: str):
        """
        Instance of API
        """
        self.host = ip_address
        self.port = port
        self.session = requests.Session()
    
    def toQuartileNumber(self, Q1: bool, Q2: bool, Q3:bool, Q4:bool, Qnan:bool) -> int:
        """
        get Quartile Number from bool values.
        """
        bitQ1, bitQ2, bitQ3, bitQ4, bitQnan = int(not Q1), int(not Q2), int(not Q3), int(not Q4), int(not Qnan)
        return bitQnan * 2**0 + bitQ4 * 2**1 + bitQ3 * 2**2 + bitQ2 * 2**3 + bitQ1 * 2**4

    #pubmed
    def find_title(self, text: str, threshold: int = 0, default: str = None, 
                   filter_quartile: int = 0, filter_country: str = None) -> str:
        """
        Search Title top-1 comparison with max score.
        """
        URL = f'http://{self.host}:{self.port}/pubmed/title/{quote(text)}?size=1&filter_quartile={str(filter_quartile)}'
        if filter_country is not None:
            URL += f'&filter_country={quote(filter_country)}'
        response = self.session.get(URL)
        if response.status_code == 200:
            result = orjson.loads(response.content)
            max_score = self.__float(result.get('max_score'))
            if max_score >= threshold:
                try:
                    return result.get('hits')[0].get('_source').get('MedlineCitation').get('Article').get('ArticleTitle', default), result.get('hits')[0].get('_id', default)
                except Exception:
                    return default, default
        return default, default
    
    #pubmed
    def find_titles_top5(self, text: str, threshold: int = 0, default: str = None) -> list:
        """
        Search Title top-5 comparisons ordered by score DESC.
        """
        URL = f'http://{self.host}:{self.port}/pubmed/title/{quote(text)}'
        response = self.session.get(URL)
        if response.status_code == 200:
            result = orjson.loads(response.content)
            max_score = self.__float(result.get('max_score'))
            if max_score >= threshold:
                retlist = []
                array = result.get('hits', [])
                for hit in array:
                    try:
                        score = self.__float(hit.get('_score'))
                        if score >= threshold:
                            title = hit.get('_source').get('MedlineCitation', {}).get('Article', {}).get('ArticleTitle', default)
                            if title not in retlist:
                                retlist.append(title)
                    except Exception:
                        return []
                return retlist
        return []

    #pubmed
    def pubmed_articles(self, text: str, pageNumber: int, size:int=20, default = None,
                     filter_quartile: int = 0, filter_country: str = None):
        """
        Default return is None. 
        """
        URL = f'http://{self.host}:{self.port}/pubmed/articles/{quote(text)}?size={str(size)}&page={str(pageNumber)}&filter_quartile={str(filter_quartile)}'
        if filter_country is not None:
            URL += f'&filter_country={quote(filter_country)}'
        response = self.session.get(URL)
        if response.status_code == 200:
            return orjson.loads(response.content)
        return default

    #PMC
    def pmc_articles(self, text: str, pageNumber: int, size:int=20, default = None,
                     filter_quartile: int = 0, filter_country: str = None):
        """
        Default return is None. 
        """
        URL = f'http://{self.host}:{self.port}/pmc/articles/{quote(text)}?size={str(size)}&page={str(pageNumber)}&filter_quartile={str(filter_quartile)}'
        if filter_country is not None:
            URL += f'&filter_country={quote(filter_country)}'
        response = self.session.get(URL)
        if response.status_code == 200:
            return orjson.loads(response.content)
        return default
    
    #PMC
    def aggs_date_year(self, default = None) -> list:
        """
        Default return is None. 
        """
        URL = f'http://{self.host}:{self.port}/pmc/count/last30years'
        response = self.session.get(URL)
        if response.status_code == 200:
            return orjson.loads(response.content)
        return default
    
    #PMC
    def get_byid(self, text: str, default = None):
        """
        Default return is None. 
        """
        URL = f'http://{self.host}:{self.port}/pmc/article/{quote(text)}'
        response = self.session.get(URL)
        if response.status_code == 200:
            return orjson.loads(response.content)         
        return default
    
    def pd_Series_get_byid(self, text: str, default = None):
        """
        return pandas.DataFrame if ModelPMC() else return Default is None.
        """
        model = self.__get_pmc_model(self.get_byid(text))
        if model is not None:
            return model.dataframe_dict_1d()
        return default

    #pmc
    def pd_DataFrame_pmc_articles(self, text: str, pageNumber: int, size:int=20, default = None,
                                  filter_quartile: int = 0, filter_country: str = None):
        content = self.pmc_articles(text, pageNumber, size, filter_quartile=filter_quartile, filter_country=filter_country)
        if content:
            count = int(content.get('total', {}).get('value', 0))
            array = []
            for hit in content.get('hits', []):
                source_dict = hit.get('_source', {})
                model = self.__get_pmc_model(source_dict)
                if model is not None:
                    array.append(model.dataframe_dict_1d())
            return array, count
        return default, 0
    
    #pubmed
    def pd_DataFrame_Pubmed_articles(self, text: str, pageNumber: int, size:int=20, default = None,
                                  filter_quartile: int = 0, filter_country: str = None):
        content = self.pubmed_articles(text, pageNumber, size, filter_quartile=filter_quartile, filter_country=filter_country)
        if content:
            count = int(content.get('total', {}).get('value', 0))
            array = []
            for hit in content.get('hits', []):
                source_dict = hit.get('_source', {})
                model = self.__get_pubmed_model(source_dict)
                if model is not None:
                    array.append(model.dataframe_dict_1d())
            return array, count
        return default, 0
    
    def __get_pmc_model(self, source: dict, default = None):
        """
        return ModelPMC() if parse_obj_as(source) else return Default is None.
        The `__get_class_validated_from_source` method takes a dictionary `source` representing a PMC article 
        and returns an instance of the `ModelPMC` class that is populated with the values from the `source` dictionary. 
        If the `source` dictionary cannot be parsed into a valid `ModelPMC` instance, 
        the method returns the value of the `default` parameter, which is set to `None` by default.

        """
        try:
            return ModelPMC(
                journal = parse_obj_as(ModelPMC.JournalModel, source.get('journal')),
                PMID = source.get('PMID', 0),
                PMCID = source.get('PMCID'),
                DOI = source.get('DOI'),
                Title = source.get('Title'),
                Author = parse_obj_as(list[ModelPMC.AuthorModel], source.get('Author')),
                date = parse_obj_as(ModelPMC.DateModel, source.get('date')),
                date_epub = source.get('date_epub'),
                Abstract = source.get('Abstract'),
                Text_full = source.get('Text_full'),
                Glossary = parse_obj_as(list[ModelPMC.DefinitionModel], source.get('Glossary')),
                Categories = source.get('Categories'),
                Affiliations = parse_obj_as(list[ModelPMC.InstituteModel], source.get('Affiliations')),
                book = parse_obj_as(ModelPMC.BookModel, source.get('book')),
                Ref_ids = parse_obj_as(list[ModelPMC.PubmedidModel], source.get('Ref_ids')),
                References = parse_obj_as(list[ModelPMC.PaperModel], source.get('References'))
            )
        except Exception:
            return default
        
    def __get_pubmed_model(self, source: dict, default = None):
        """
        return ModelPubMedLine() if parse_obj_as(source) else return Default is None.
        The method takes a dictionary `source` representing a Pubmed article 
        and returns an instance of the `ModelPubMedLine` class that is populated with the values from the `source` dictionary. 
        If the `source` dictionary cannot be parsed into a valid `ModelPubMedLine` instance, 
        the method returns the value of the `default` parameter, which is set to `None` by default.

        """
        try:
            MedlineCitationDict = source.get('MedlineCitation')
            ArticleDict = MedlineCitationDict.get('Article')
            JournalDict = ArticleDict.get('Journal')
            JournalIssueDict = JournalDict.get('JournalIssue', {})
            PubmedDataDict = source.get('PubmedData')
            return ModelPubMedLine(
                MedlineCitation = ModelPubMedLine.MedlineCitationModel(
                    PMID = MedlineCitationDict.get('PMID'),
                    DateRevised = parse_obj_as(ModelPubMedLine.MedlineCitationModel.DateRevisedModel, MedlineCitationDict.get('DateRevised', {})),
                    Article = ModelPubMedLine.MedlineCitationModel.ArticleModel(
                        Journal = ModelPubMedLine.MedlineCitationModel.ArticleModel.JournalModel(
                            ISOAbbreviation = JournalDict.get('ISOAbbreviation'),
                            Title = JournalDict.get('Title'),
                            JournalIssue = ModelPubMedLine.MedlineCitationModel.ArticleModel.JournalModel.JournalIssueModel(
                                Volume = JournalIssueDict.get('Volume', ''),
                                Issue = JournalIssueDict.get('Issue', ''),
                                PubDate = parse_obj_as(ModelPubMedLine.MedlineCitationModel.ArticleModel.JournalModel.JournalIssueModel.PubDateModel, JournalIssueDict.get('PubDate', {})),
                            ),
                            ISSN_Print = JournalDict.get('ISSN_Print'),
                            ISSN_Electronic = JournalDict.get('ISSN_Electronic'),
                        ),
                        ArticleTitle = ArticleDict.get('ArticleTitle'),
                        Pagination = ArticleDict.get('Pagination', ''),
                        ELocationID = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.ArticleModel.ELocationIDModel], ArticleDict.get('ELocationID', [])),
                        AbstractText = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.ArticleModel.AbstractTextModel], ArticleDict.get('AbstractText', [])),
                        CopyrightInformation = ArticleDict.get('CopyrightInformation', ''),
                        AuthorList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.ArticleModel.AuthorModel], ArticleDict.get('AuthorList', [])),
                        PublicationTypeList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.ArticleModel.PublicationTypeModel], ArticleDict.get('PublicationTypeList', [])),
                    ),
                    MedlineJournalInfo = parse_obj_as(ModelPubMedLine.MedlineCitationModel.MedlineJournalInfoModel, MedlineCitationDict.get('MedlineJournalInfo')),
                    MeshHeadingList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.MeshHeadingModel], MedlineCitationDict.get('MeshHeadingList', [])),
                    KeywordList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.KeywordModel], MedlineCitationDict.get('KeywordList', [])),
                    CommentsCorrectionsList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.CommentsCorrectionsModel], MedlineCitationDict.get('CommentsCorrectionsList', [])),
                    ChemicalList = parse_obj_as(list[ModelPubMedLine.MedlineCitationModel.ChemicalModel], MedlineCitationDict.get('ChemicalList', [])),
                ),
                PubmedData = ModelPubMedLine.PubmedDataModel(
                    History = parse_obj_as(list[ModelPubMedLine.PubmedDataModel.PubMedPubDateModel], PubmedDataDict.get('History')),
                    ArticleIdList = parse_obj_as(list[ModelPubMedLine.PubmedDataModel.ArticleIdModel], PubmedDataDict.get('ArticleIdList')),
                    ReferenceList = parse_obj_as(list[ModelPubMedLine.PubmedDataModel.ReferenceModel], PubmedDataDict.get('ReferenceList', [])),
                ),
                DateEntrez = source.get('DateEntrez')
            )
        except Exception:
            return default

    def __float(self, x):
        """
        The `__float` method is a private helper method that takes a value `x` and attempts to parse it as a floating-point number. 
        If parsing succeeds, the method returns the float value of `x`. If parsing fails, the method returns 0. 
        The double underscores in the method name indicate that it is meant to be a private method.
        """
        try:
            return float(x)
        except Exception:
            return 0


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Iterable, List
from collections import deque

from .chat import Chat


class ChatMenu:
    
    def __init__(self, chats: Iterable[Chat]=None):
        if chats is None:
            self.chats = deque()
        else:
            self.chats = deque(chats)
        return
    
    def __getitem__(self, item: int) -> Chat:
        return self.chats[item]
    
    def to_front(self) -> List[dict]:
        return [chat.to_front() for chat in self.chats]


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Any

import requests

from .containers.rdict import RDict


def method_logger(func):
    """
    Special decorator for logging RAI ChatBot methods.
    """
    
    def start_log(func_name, args, kwargs):
        strings = [
            f"[INFO]: EXECUTING: {func.__name__}",
            f"[INFO]: Args: {args}",
            f"[INFO]: Kwargs: {kwargs}"
        ]
        print("\n".join(strings))
        return
    
    def end_log(func_name, args, kwargs):
        strings = [
            f"[INFO]: FINISHED: {func.__name__}",
            f"[INFO]: Args: {args}",
            f"[INFO]: Kwargs: {kwargs}"
        ]
        print("\n".join(strings))
        return
        
    def safe_log(logger, obj, func_name, args, kwargs):
        try:
            if obj.log:
                logger(func_name, args, kwargs)
        except Exception:
            pass
        return
    
    def wrapper(*args, **kwargs):
        obj = args[0]
        safe_log(start_log, obj, func.__name__, args, kwargs)
        res = func(*args, **kwargs)
        safe_log(end_log, obj, func.__name__, args, kwargs)
        return res
    
    return wrapper


def retry(n_retries: int=5):
    """
    Retrying decorator function.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            n = n_retries
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    n -= 1
                    if n <= 0:
                        raise
        return wrapper
    return decorator


def to_rdict(obj: object, recursive: bool=True) -> RDict:
    """
    Converts JSON-like object into RDict.
    Args:
        obj[object]: JSON-like object to convert.
        recursive[bool]: Whether to convert all nested JSON-like objects. Default: True.
    Returns:
        RDict
    """
    if not (isinstance(obj, list) or isinstance(obj, dict)):
        return obj
    # Iterate over list elements, if convertion is recursive
    elif isinstance(obj, list):
        if not recursive:
            return obj
        return [to_rdict(elem) for elem in obj]
    # Iterate over dict elements, if convertion is recursive
    elif isinstance(obj, dict):
        if not recursive:
            return RDict(obj)
        return RDict({key: to_rdict(value) for key, value in obj.items()})

def request_openai(**kwargs) -> dict:
    """
    Sends request to OpenAI API endpoint in order to obtain response.
    Is mainly used for chat-like interactions with GPT API.
    """
    url = "https://api.openai.com/v1/chat/completions"
    api_key = kwargs.pop("api_key", None)
    headers = {"Authorization": f"Bearer {api_key}"}
    r = requests.post(url, headers=headers, json=kwargs)
    return r.json()


from pydantic import BaseModel
from pydantic import parse_obj_as
# Python >= 3.9 version required

#PMC
class ModelPMC(BaseModel):
    """
    Serializable model of /index of OpenSearch
    """
    class JournalModel(BaseModel):
        nlm: str = ''
        longname: str = ''
        ISSNP: str = ''
        ISSNE: str = ''

    class DateModel(BaseModel):
        year: int
        month: int = 0
        day: int = 0

    class AuthorModel(BaseModel):
        fullname: str = ''
        initials: str = ''

    class DefinitionModel(BaseModel):
        definition: str = ''

    class InstituteModel(BaseModel):
        institute: str = ''

    class BookModel(BaseModel):
        volume: str = ''
        issue: str = ''
        first_page: str = ''
        last_page: str = ''

    class PubmedidModel(BaseModel):
        pubmed_id: int

    class PaperModel(BaseModel):
        paper: str = ''

    # @properties
    journal: JournalModel
    PMID: int = 0
    PMCID: str
    DOI: str = ''
    Title: str = ''
    Author: list[AuthorModel] = []
    date: DateModel
    date_epub: str
    Abstract: str = ''
    Text_full: str = ''
    Glossary: list[DefinitionModel] = []
    Categories: str = ''
    Affiliations: list[InstituteModel] = []
    book: BookModel
    Ref_ids: list[PubmedidModel] = []
    References: list[PaperModel] = []

    def dataframe_dict_1d(self):
        __list__Author_fullname = []
        __list__Author_initials = []
        __list__Glossary = []
        __list__Affilations = []
        __list__Refids = []
        __list__References = []
        
        for author in self.Author:
            __list__Author_fullname.append(author.fullname)
            __list__Author_initials.append(author.initials)
        for glossary in self.Glossary:
            __list__Glossary.append(glossary.definition)
        for affiliation in self.Affiliations:
            __list__Affilations.append(affiliation.institute)
        for refid in self.Ref_ids:
            __list__Refids.append(str(refid.pubmed_id))
        for reference in self.References:
            __list__References.append(reference.paper)

        return {
            "PMID": self.PMID,
            "PMCID": self.PMCID,
            "Title": self.Title,
            "Categories": self.Categories,
            "date_epub": self.date_epub,
            "Abstract": self.Abstract,
            "Author.initials": ", ".join(__list__Author_initials),
            "Author.fullname": ", ".join(__list__Author_fullname),
            "Affiliations": "; ".join(__list__Affilations),
            "Text_full": self.Text_full,
            "Glossary": "; ".join(__list__Glossary),
            "journal.nlm": self.journal.nlm,
            "journal.longname": self.journal.longname,
            "journal.ISSNP": self.journal.ISSNP,
            "journal.ISSNE": self.journal.ISSNE,
            "DOI": self.DOI,
            "book.volume": self.book.volume,
            "book.issue": self.book.issue,
            "book.first_page": self.book.first_page,
            "book.last_page": self.book.last_page,
            "date.year": self.date.year,
            "date.month": self.date.month,
            "date.day": self.date.day,
            "Ref_ids": ",".join(__list__Refids),
            "References": "\n".join(__list__References),
        }
    
    @classmethod
    def from_source(self, source: dict, default = None):
        try:
            return self(
                journal = parse_obj_as(self.JournalModel, source.get('journal')),
                PMID = source.get('PMID', 0),
                PMCID = source.get('PMCID'),
                DOI = source.get('DOI'),
                Title = source.get('Title'),
                Author = parse_obj_as(list[self.AuthorModel], source.get('Author')),
                date = parse_obj_as(self.DateModel, source.get('date')),
                date_epub = source.get('date_epub'),
                Abstract = source.get('Abstract'),
                Text_full = source.get('Text_full'),
                Glossary = parse_obj_as(list[self.DefinitionModel], source.get('Glossary')),
                Categories = source.get('Categories'),
                Affiliations = parse_obj_as(list[self.InstituteModel], source.get('Affiliations')),
                book = parse_obj_as(self.BookModel, source.get('book')),
                Ref_ids = parse_obj_as(list[self.PubmedidModel], source.get('Ref_ids')),
                References = parse_obj_as(list[self.PaperModel], source.get('References'))
            )
        except Exception:
            return default

#PubMedLine
class ModelPubMedLine(BaseModel):
    """
    Serializable model: Pubmed/MEDLINE
    """
    class MedlineCitationModel(BaseModel):
        class DateRevisedModel(BaseModel):
            Year: str
            Month: str
            Day: str
        class ArticleModel(BaseModel):
            class JournalModel(BaseModel):
                class JournalIssueModel(BaseModel):
                    class PubDateModel(BaseModel):
                        Year: str = ''
                        Month: str = ''
                        Day: str = ''
                    Volume: str = ''
                    Issue: str = ''
                    PubDate: PubDateModel
                ISOAbbreviation: str
                Title: str
                JournalIssue: JournalIssueModel
                ISSN_Print: str = ''
                ISSN_Electronic: str = ''
            class AuthorModel(BaseModel):
                class AffiliationModel(BaseModel):
                    Affiliation: str = ''
                LastName: str = ''
                ForeName: str = ''
                Initials: str = ''
                AffiliationInfo: list[AffiliationModel] = []
            class ELocationIDModel(BaseModel):
                EIdType: str = ''
                text: str = ''
            class PublicationTypeModel(BaseModel):
                UI: str
                text: str
            class AbstractTextModel(BaseModel):
                Label: str = ''
                NlmCategory: str = ''
                text: str
            Journal: JournalModel
            ArticleTitle: str
            Pagination: str = ''
            ELocationID: list[ELocationIDModel] = []
            AbstractText: list[AbstractTextModel] = []
            CopyrightInformation: str = ''
            AuthorList: list[AuthorModel] = []
            PublicationTypeList: list[PublicationTypeModel]
        class MedlineJournalInfoModel(BaseModel):
            MedlineTA: str
            NlmUniqueID: str
            Country: str = ''
            ISSNLinking: str = ''
        class MeshHeadingModel(BaseModel):
            UI: str = ''
            text: str = ''
        class KeywordModel(BaseModel):
            text: str = ''
        class CommentsCorrectionsModel(BaseModel):
            RefType: str = ''
            RefSource: str = ''
            PMID: str
        class ChemicalModel(BaseModel):
            class NameOfSubstanceModel(BaseModel):
                UI: str
                text: str
            RegistryNumber: str
            NameOfSubstance: NameOfSubstanceModel
        PMID: int
        DateRevised: DateRevisedModel
        Article: ArticleModel
        MedlineJournalInfo: MedlineJournalInfoModel
        MeshHeadingList: list[MeshHeadingModel] = []
        KeywordList: list[KeywordModel] = []
        CommentsCorrectionsList: list[CommentsCorrectionsModel] = []
        ChemicalList: list[ChemicalModel] = []
    class PubmedDataModel(BaseModel):
        class PubMedPubDateModel(BaseModel):
            Year: str
            Month: str
            Day: str
            Hour: str
            Minute: str
            PubStatus: str = ''
        class ArticleIdModel(BaseModel):
            IdType: str
            text: str
        class ReferenceModel(BaseModel):
            class ArticleIdModel(BaseModel):
                IdType: str
                text: str
            Citation: str
            ArticleIdList: list[ArticleIdModel]
        History: list[PubMedPubDateModel]
        ArticleIdList: list[ArticleIdModel]
        ReferenceList: list[ReferenceModel] = []
    MedlineCitation: MedlineCitationModel
    PubmedData: PubmedDataModel
    DateEntrez: str

    @property
    def PMCID(self, default: str = ''):
        for articleID in self.PubmedData.ArticleIdList:
            if articleID.IdType == 'pmc':
                return articleID.text
        return default
    
    @property
    def DOI(self, default: str = ''):
        for articleID in self.PubmedData.ArticleIdList:
            if articleID.IdType == 'doi':
                return articleID.text
        return default
    
    @property
    def Categories(self):
        return " ".join([pubType.text for pubType in self.MedlineCitation.Article.PublicationTypeList])
    
    @property
    def Abstract(self):
        return "\n".join([it.text for it in self.MedlineCitation.Article.AbstractText])

    def dataframe_dict_1d(self):
        __list__Author_fullname = []
        __list__Author_initials = []
        __list__Glossary = []
        __list__Affilations = []
        __list__Refids = []
        __list__References = []
        
        for author in self.MedlineCitation.Article.AuthorList:
            __list__Author_fullname.append(author.ForeName + ' ' + author.LastName)
            __list__Author_initials.append(author.LastName + ' ' + author.Initials)
            for aff in author.AffiliationInfo:
                if aff.Affiliation not in __list__Affilations:
                    __list__Affilations.append(aff.Affiliation)
        for keyword in self.MedlineCitation.KeywordList:
            __list__Glossary.append(keyword.text)
        for ref in self.PubmedData.ReferenceList:
            pmid = '0'
            for articleID in ref.ArticleIdList:
                if articleID.IdType == 'pubmed':
                    pmid = articleID.text
                    break
            if pmid != '0':
                __list__Refids.append(pmid)
            else:
                __list__References.append(ref.Citation.strip())

        return {
            "PMID": self.MedlineCitation.PMID,
            "PMCID": self.PMCID,
            "Title": self.MedlineCitation.Article.ArticleTitle,
            "Categories": self.Categories,
            "date_epub": self.DateEntrez,
            "Abstract": self.Abstract,
            "Author.initials": ", ".join(__list__Author_initials),
            "Author.fullname": ", ".join(__list__Author_fullname),
            "Affiliations": "; ".join(__list__Affilations),
            "Text_full": None,
            "Glossary": "; ".join(__list__Glossary),
            "journal.nlm": self.MedlineCitation.Article.Journal.ISOAbbreviation,
            "journal.longname": self.MedlineCitation.Article.Journal.Title,
            "journal.ISSNP": self.MedlineCitation.Article.Journal.ISSN_Print,
            "journal.ISSNE": self.MedlineCitation.Article.Journal.ISSN_Electronic,
            "DOI": self.DOI,
            "book.volume": self.MedlineCitation.Article.Journal.JournalIssue.Volume,
            "book.issue": self.MedlineCitation.Article.Journal.JournalIssue.Issue,
            "book.first_page": self.MedlineCitation.Article.Pagination,
            "book.last_page": '',
            "date.year": self.MedlineCitation.Article.Journal.JournalIssue.PubDate.Year,
            "date.month": self.MedlineCitation.Article.Journal.JournalIssue.PubDate.Month,
            "date.day": self.MedlineCitation.Article.Journal.JournalIssue.PubDate.Day,
            "Ref_ids": ",".join(__list__Refids),
            "References": "\n".join(__list__References),
        }
    
    @classmethod
    def from_source(self, source: dict, default = None):
        try:
            MedlineCitationDict = source.get('MedlineCitation')
            ArticleDict = MedlineCitationDict.get('Article')
            JournalDict = ArticleDict.get('Journal')
            JournalIssueDict = JournalDict.get('JournalIssue', {})
            PubmedDataDict = source.get('PubmedData')
            return self(
                MedlineCitation = self.MedlineCitationModel(
                    PMID = MedlineCitationDict.get('PMID'),
                    DateRevised = parse_obj_as(self.MedlineCitationModel.DateRevisedModel, MedlineCitationDict.get('DateRevised', {})),
                    Article = self.MedlineCitationModel.ArticleModel(
                        Journal = self.MedlineCitationModel.ArticleModel.JournalModel(
                            ISOAbbreviation = JournalDict.get('ISOAbbreviation'),
                            Title = JournalDict.get('Title'),
                            JournalIssue = self.MedlineCitationModel.ArticleModel.JournalModel.JournalIssueModel(
                                Volume = JournalIssueDict.get('Volume', ''),
                                Issue = JournalIssueDict.get('Issue', ''),
                                PubDate = parse_obj_as(self.MedlineCitationModel.ArticleModel.JournalModel.JournalIssueModel.PubDateModel, JournalIssueDict.get('PubDate', {})),
                            ),
                            ISSN_Print = JournalDict.get('ISSN_Print'),
                            ISSN_Electronic = JournalDict.get('ISSN_Electronic'),
                        ),
                        ArticleTitle = ArticleDict.get('ArticleTitle'),
                        Pagination = ArticleDict.get('Pagination', ''),
                        ELocationID = parse_obj_as(list[self.MedlineCitationModel.ArticleModel.ELocationIDModel], ArticleDict.get('ELocationID', [])),
                        AbstractText = parse_obj_as(list[self.MedlineCitationModel.ArticleModel.AbstractTextModel], ArticleDict.get('AbstractText', [])),
                        CopyrightInformation = ArticleDict.get('CopyrightInformation', ''),
                        AuthorList = parse_obj_as(list[self.MedlineCitationModel.ArticleModel.AuthorModel], ArticleDict.get('AuthorList', [])),
                        PublicationTypeList = parse_obj_as(list[self.MedlineCitationModel.ArticleModel.PublicationTypeModel], ArticleDict.get('PublicationTypeList', [])),
                    ),
                    MedlineJournalInfo = parse_obj_as(self.MedlineCitationModel.MedlineJournalInfoModel, MedlineCitationDict.get('MedlineJournalInfo')),
                    MeshHeadingList = parse_obj_as(list[self.MedlineCitationModel.MeshHeadingModel], MedlineCitationDict.get('MeshHeadingList', [])),
                    KeywordList = parse_obj_as(list[self.MedlineCitationModel.KeywordModel], MedlineCitationDict.get('KeywordList', [])),
                    CommentsCorrectionsList = parse_obj_as(list[self.MedlineCitationModel.CommentsCorrectionsModel], MedlineCitationDict.get('CommentsCorrectionsList', [])),
                    ChemicalList = parse_obj_as(list[self.MedlineCitationModel.ChemicalModel], MedlineCitationDict.get('ChemicalList', [])),
                ),
                PubmedData = self.PubmedDataModel(
                    History = parse_obj_as(list[self.PubmedDataModel.PubMedPubDateModel], PubmedDataDict.get('History')),
                    ArticleIdList = parse_obj_as(list[self.PubmedDataModel.ArticleIdModel], PubmedDataDict.get('ArticleIdList')),
                    ReferenceList = parse_obj_as(list[self.PubmedDataModel.ReferenceModel], PubmedDataDict.get('ReferenceList', [])),
                ),
                DateEntrez = source.get('DateEntrez')
            )
        except Exception:
            return default


from setuptools import setup, find_packages

setup(name="redemption-ai",
      version="0.0.6",
      description="AI-powered (OpenAI API) tools for chat-like interactions and personal AI assistance.",
      packages=find_packages(include=["RAI", "RAI.*"]),
      author="Ausar686",
      author_email='glebyushkov@mail.ru',
      install_requires=[
        "numpy==1.25.2",
        "pandas==2.1.0",
        "requests==2.31.0",
        "urllib3==2.0.4"
      ])


# Created by: Ausar686
# https://github.com/Ausar686

from typing import Any, Union
import hashlib
import json
import os
import re

from .token_counter import TokenCounter
from .text_summarizer import TextSummarizer
from ..chat import  Chat, Message
from ..containers import RDict 
from ..profile import Profile
from ..utils import method_logger, retry, to_rdict, request_openai


class ChatBot:
    """
    RAI class for chat-like interaction with OpenAI API using customizable set of actors.
    """
    
    _defaults = RDict({
        "openai": RDict({
            "api_key": None,
            "model": "gpt-3.5-turbo",
            "temperature": 0,
            "stream": False,
            "n": 1,
        }),
        "limits": RDict({
            "gpt-3.5-turbo": 3000,
            "gpt-3.5-turbo-16k": 15000,
            "gpt-4": 7000,
            "gpt-4-32k": 31000
        }),
        "upgrades": RDict({
            "gpt-3.5-turbo": "gpt-3.5-turbo-16k",
            "gpt-4": "gpt-4-32k"
        }),
        "downgrades": RDict({
            "gpt-3.5-turbo-16k": "gpt-3.5-turbo",
            "gpt-4-32k": "gpt-4"
        }),
        "chat": RDict({
            "username": "DefaultUser",
            "bot_name": "DefaultBot"
        }),
        "text_summarizer": RDict({
            "model": "gpt-3.5-turbo",
            "n_words": 50
        }),
        "token_counter": RDict({
            "model": "gpt-3.5-turbo"
        }),
        "qagpt": RDict({
            "model": "gpt-3.5-turbo",
            "temperature": 0,
            "stream": False,
            "n": 1
        })
    })
    
    _runtime_modes_available = ["console", "app"]
    
    # Errors initialization
    _api_key_error = """
        OpenAI API key must be provided in one of two ways.
        1. Via .json configuration file:
            "openai": {
                "api_key": "YOUR_KEY_HERE",
                ...
            }
        2. Via environment variable "OPENAI_API_KEY".
    """
    
    _last_message_len_error = "Last message is too long to proceed."
    
    _n_answers_error = "Several answers option is not implemented yet."
    
    _mode_error = f"Wrong mode value. Available modes are: {_runtime_modes_available}"
        
    def __init__(
        self,
        config_path: str,
        actors_config_path: str=None,
        functions_config_path: str=None,
        *,
        mode: str="console",
        profile: Profile=None,
        chat: Chat=None,
        log: bool=False) -> None:
        """
        Initializes ChatBot instance using 2 config files.
        """
        # Initialize utils for uploading data from '.ini' config file
        self.log = log
        self.name = None
        self.config_dir = None
        self.parameters = RDict()
        self.actors = RDict()
        self._synced_actors = []
        self._runtime_mode = mode
        # Load config options from '.json' config file
        self.from_config(config_path)
        # Set OpenAI API parameters
        self.set_openai_parameters()
        # Setup actors from config
        self.actors_from_config(actors_config_path)
        # Setup usable functions from config.
        self.functions_from_config(functions_config_path)
        # Initialize dialog attributes
        self.init_messages(profile, chat)
        # Validate initialization
        self.validate_init()
        return
    
    @method_logger
    def from_config(self, config_path: str=None) -> None:
        """
        Runs instance initialization from .json configuration file.
        """
        if config_path is None:
            return
        if self.config_dir is None:
            self.config_path = os.path.split(os.path.abspath(config_path))[0]
        with open(config_path, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
        for section in data.keys():
            section_data = data.get(section)
            self.process_section(section, section_data)
        return
    
    @method_logger
    def process_section(self, section: str, section_data: Union[dict, str]) -> None:
        """
        Processes one section of the configuration file.
        """
        # Initialize from parent config file.
        if section == "root":
            self.from_config(section_data)
            return
        # Set bot name
        elif section == "name":
            self.name = section_data
            return
        # Set attribute as a key in self.parameters dict
        # if it's not present yet
        if not hasattr(self, section):
            self.parameters[section] = RDict()
        for option, value in section_data.items():
            self.process_option(section, option, value)
        return
    
    @method_logger
    def process_option(self, section: str, option: str, value: Any=None) -> None:
        """
        Processes one option in a section of the configuration file.
        """
        self.parameters[section][option] = value
        return
    
    @method_logger
    def actors_from_config(self, actors_config_path: str=None) -> None:
        """
        Initializes ChatBot actors from .json configuration file.
        """
        if actors_config_path is None:
            # Set only default actors
            self.set_default_actors()
            return
        with open(actors_config_path, "r", encoding="utf-8") as json_file:
            actors_json = json.load(json_file)
        for actor in actors_json:
            self.set_actor(actor)
        # Set default actors if they are not set yet.
        self.set_default_actors()
        return

    @method_logger
    def set_actor(self, actor: dict) -> None:
        """
        Sets an actor to bot, using data in dictionary, obtained from .json file.
        """
        name = actor.get("name")
        class_name = actor.get("class")
        kwargs = actor.get("params")
        sync = actor.get("sync_models")
        exec(f"from RAI import {class_name}")
        exec(f"self.actors[{name}] = {class_name}(**{kwargs})")
        if sync:
            self._synced_actors.append(self.actors[name])
        return
    
    @method_logger
    def set_default_actors(self):
        """
        Initializes default ChatBot actors.
        """
        if "token_counter" not in self.actors:
            self.actors.token_counter = TokenCounter(self.openai.model)
        if "summarizer" not in self.actors:
            self.actors.summarizer = TextSummarizer(self.actors.token_counter.model, self._defaults.text_summarizer.n_words)
        self._synced_actors.append(self.actors.token_counter)
        return
    
    @method_logger
    def set_api_key(self) -> None:
        """
        Sets OpenAI API key.
        """
        if not hasattr(self.openai, "api_key") or self.openai.api_key is None:
            self.openai.api_key = os.getenv("OPENAI_API_KEY")
        if self.openai.api_key is None:
            raise ValueError(self._api_key_error)
        # Set environmental variable to prevent overriding by further imports
        os.environ["OPENAI_API_KEY"] = self.openai.api_key
        return
    
    @method_logger
    def set_openai_parameters(self) -> None:
        """
        Sets all required OpenAI API parameters (excluding API key).
        """
        if not hasattr(self, "openai"):
            raise ValueError("OpenAI section is not present in configuration files.")
        for key, value in self._defaults.openai.items():
            if key == "api_key":
                self.set_api_key()
                continue
            if key not in self.openai:
                self.openai[key] = value
        return
    
    @method_logger
    def functions_from_config(self, functions_config_path: str=None) -> None:
        """
        Set functions list in JSON notation to use in OpenAI API calls.
        """
        if functions_config_path is None:
            self.set_default_functions()
            return
        with open(functions_config_path, "r", encoding="utf-8") as json_file:
            self.usable_functions = json.load(json_file)
        self._usable_functions_names = [elem["name"] for elem in self.usable_functions]
        return
    
    @method_logger
    def set_default_functions(self) -> None:
        """
        Set default list of functions (None) to use in OpenAI API calls.
        """
        self.usable_functions = None
        self._usable_functions_names = None
        return
    
    @method_logger
    def set_instruction(self) -> None:
        """
        Sets bot instruction in a string representation as an attribute.
        """
        self.instruction = "###INSTRUCTIONS###\n" + "\n\n".join([f"<{key}>: {value}" for key, value in self.instructions.items()])
        return
    
    @method_logger
    def set_system_message(self) -> None:
        """
        Sets system message in OpenAI notation for the bot as an attribute.
        """
        self.set_instruction()
        self.system_message = Message({"role": "system", "content": self.instruction, "username": "ROOT"})
        return
    
    @method_logger
    def init_messages(self, profile: Profile, chat: Chat) -> None:
        """
        Initializes dialog attributes.
        'chat' is an internal message storage (will NOT be sent to API).
        'context' is a list of messages in OpenAI notation (will be sent to API).
        """
        if profile is None:
            self.username = self._defaults["chat"]["username"]
        else:
            self.username = profile.user_data.username
        self.profile = profile
        if self.name is None:
            self.name = self._defaults["chat"]["bot_name"]
        self.set_system_message()
        if chat is None:
            chat_dct = {"username": self.username, "bot_name": self.name, "messages": [self.system_message]}
            self.chat = Chat(chat_dct)
        else:
            self.chat = chat
        self.context = [self.system_message.to_openai()]
        return
    
    @method_logger
    def validate_init(self) -> None:
        """
        Validates initialization. Use this method to raise exceptions for not implemented options.
        """
        if self.openai.n > 1:
            raise NotImplementedError(self._n_answers_error)
        if self._runtime_mode not in ["console", "app"]:
            raise NotmplementedError(self._mode_error)
        return

    @property
    def messages(self) -> list:
        """
        Returns all messages in self.chat
        """
        return self.chat.messages
    
    @property
    def last_message(self) -> Message:
        """
        Returns last message in self.chat or None if self.chats is empty.
        """
        return self.chat.last if self.chat else None
    
    def append(self, msg: Message) -> None:
        """
        Appends message both to context and to chat.
        """
        self.chat.append(msg)
        self.context.append(msg.to_openai())
        return
    
    def user_message(self, text: str) -> Message:
        """
        Converts text to Message with user data.
        """
        msg_dct = {"role": "user", "content": text, "username": self.username}
        msg = Message(msg_dct)
        return msg
    
    def bot_message(self, text: Any) -> Message:
        """
        Converts text to Message with bot data.
        """
        msg_dct = {"role": "assistant", "content": str(text), "username": self.name}
        msg = Message(msg_dct)
        return msg
    
    def sys_message(self, text: str) -> Message:
        """
        Converts text to message with system data.
        """
        msg_dct = {"role": "system", "content": text, "username": "ROOT"}
        msg = Message(msg_dct)
        return msg
    
    def add_user_message(self, text: str) -> None:
        """
        Appends user message both to context and to chat.
        """
        msg = self.user_message(text)
        self.append(msg)
        return
    
    def add_bot_message(self, text: str) -> None:
        """
        Appends bot message both to context and to chat.
        """
        msg = self.bot_message(text)
        self.append(msg)
        return
    
    def upgrade_model(self) -> None:
        """
        Upgrades token limit of the using OpenAI API model.
        """
        self.openai.model = self._defaults.upgrades[self.openai.model]
        self.set_actors_model()
        return
    
    def downgrade_model(self) -> None:
        """
        Downgrades token limit of the using OpenAI API model.
        """
        self.openai.model = self._defaults.downgrades[self.openai.model]
        self.set_actors_model()
        return
    
    def set_actors_model(self) -> None:
        """
        Syncronizes models among all required actors in instance.
        """
        for actor in self._synced_actors:
            actor.set_model(self.openai.model)
        return

    def __len__(self) -> int:
        """
        Returns the current size of the bot context in tokens.
        """
        return self.actors.token_counter.run(self.context)

    @property
    def last_message_len(self) -> int:
        """
        Returns the current size of the last message in tokens
        """
        return self.actors.token_counter.run(self.last_message)

    @property
    def size_limit(self) -> int:
        """
        Returns available token size limit for context for self.openai.model
        """
        return self._defaults.limits[self.openai.model]
    
    def verify_context(self) -> None:
        """
        Verifies, that context length is not out-of-range.
        If it is, summarizes the context and updates it.
        Updated context contains system message, summary and last message.
        Chat data is not affected by this method.
        """
        if len(self) > self.size_limit:
            # Check for huge prompt injection
            if self.last_message_len > self.size_limit:
                try:
                    self.upgrade_model()
                except KeyError:
                    self.fix_injection()
                self.verify_context()
                return
            try:
                self.upgrade_model()
            except KeyError:
                self.summarize_context()
            self.verify_context()
        return

    def summarize_context(self) -> None:
        """
        Performs context summarization. By default, is used once context length reaches the limit.
        """
        summary = self.actors.summarizer.run(self.context[1:-1])
        msg = self.sys_message(summary)
        msg_list = [self.system_message, msg, self.last_message]
        self.context = [elem.to_openai() for elem in msg_list]
        try:
            self.downgrade_model()
        except KeyError:
            pass
        return
    
    def fix_injection(self) -> None:
        """
        Fixes huge message injection by popping it from context.
        Also writes an error log message.
        """
        error_text = self._last_message_len_error
        error_msg = self.bot_message(error_text)
        self.chat.append(error_msg)
        self.context.pop()
        return
    
    @retry(5)
    def get_completion(self) -> RDict:
        """
        Sends context to OpenAI API and receives response from it as a completion.
        """
        json_data = request_openai(messages=self.context, **self.openai)
        completion = to_rdict(json_data)
        return completion
    
    @retry(5)
    def fget_completion(self) -> RDict:
        """
        Sends context to OpenAI API and receives response from it as a completion.
        Response CAN contain a function call from self.usable_functions.
        """
        if self.usable_functions is None:
            return self.get_completion()
        json_data = request_openai(
            functions=self.usable_functions,
            messages=self.context,
            **self.openai)
        completion = to_rdict(json_data)
        return completion
    
    @staticmethod
    def completion_to_openai_message_list(completion: RDict) -> list:
        """
        Converts OpenAI Completion (RDict) to a list of choices.
        """
        lst = [choice.message for choice in completion.choices]
        return lst
    
    @staticmethod
    def openai_message_list_to_dict(lst: list) -> RDict:
        """
        Converts list of messages to a RDict:
        {"role": role, "content": content}
        """
        # [TODO]: Add proper processing of several generated variants
        return lst[0]

    @staticmethod
    def remove_numeration(text: str) -> str:
        """
        Removes numeration from the text, thus converting numerated lists into plain text.
        """
        re_text = re.sub(r"(\n+?)\d+\.\s*(.*?)", r"\n\2", text)
        return re_text
    
    def call_from_completion(self, completion: RDict) -> Any:
        """
        Calls function from OpenAI Completion.
        Raises ValueError if function is not presented in self.usable_functions
        """
        # [TODO]: Add processing of multiple choices.
        message = completion.choices[0].message
        call_data = message.function_call
        func_name = call_data.name
        if func_name not in self._usable_functions_names:
            raise ValueError(f"Function {func_name} is not available to call.")
        arg_string = call_data.arguments
        kwargs = json.loads(arg_string)
        content = eval(f"self.{func_name}(**{kwargs})")
        msg = self.bot_message(content)
        return msg
    
    def completion_to_message(self, completion: RDict) -> Message:
        """
        Converts OpenAI Completion into a Message
        """
        msg_lst = self.completion_to_openai_message_list(completion)
        msg_dct = self.openai_message_list_to_dict(msg_lst)
        msg_dct["content"] = self.remove_numeration(msg_dct["content"])
        msg = Message(msg_dct)
        return msg
    
    def process_completion(self, completion: RDict) -> Any:
        """
        Processes the completion.
        If function call is required, calls it and returns the result.
        """
        # [TODO]: Add processing of multiple choices.
        finish_reason = completion.choices[0].finish_reason
        if finish_reason == "stop":
            return self.completion_to_message(completion)
        elif finish_reason == "function_call":
            return self.call_from_completion(completion)
        else:
            raise ValueError(f"Invalid value of 'finish_reason': {finish_reason}")
    
    def get_answer(self) -> Message:
        """
        Full pipeline of answer obtaining from OpenAI API.
        """
        completion = self.get_completion()
        msg = self.process_completion(completion) 
        return msg

    def fget_answer(self) -> Message:
        """
        Full pipeline of answer obtaining from OpenAI API.
        """
        # [TODO]: Add proper exception handling
        try:
            completion = self.fget_completion()
            msg = self.process_completion(completion)
            return msg
        except Exception:
            return self.get_answer()

    @property
    def last_message_fstring(self) -> str:
        """
        Returns last message as a formatted string.
        """
        return f"[{self.last_message.username}]: {self.last_message.content}"

    @property    
    def is_over(self) -> bool:
        """
        A property, that represents, whether the dialogue is ended by user.
        """
        return not bool(self.last_message.content)

    def to_txt(self, path: str) -> None:
        """
        Writes chat in .txt file in the format below.
        Example:
            [Bot]: How can I assist you?
            [User]: I want to become rich.
            [Bot]: ...
            ..........
        """
        msg_lst =  [f"{msg.username}: {msg.content}" for msg in self.messages[1:]]
        text = "\n".join(msg_lst)
        with open(path, "w", encoding="utf-8") as file:
            file.write(text)
        return

    def clear(self) -> None:
        """
        Cleans the context for bot reusing
        NOTE: This method does NOT clean the entire chat (self.chat)
        """
        self.context.clear()
        self.context.append(self.system_message.to_openai())
        return
    
    def __getattr__(self, attr: str) -> Any:
        """
        Provides attribute access to self.parameters
        Note: Do not use default names as parameters attributes, as it will not work properly with __getattr__
        """
        if attr in self.parameters:
            return self.parameters.get(attr)
        raise AttributeError(f"Attribute {attr} does not exist.")

    # Here are some beta-feature methods, which will be useful in future updates.

    def hash_button_option(self, option: str) -> str:
        prefix = f"{self.username}~~~{self.name}"
        hashed_prefix = hashlib.sha256(str.encode(prefix)).hexdigest()
        hashed_string = f"{hashed_prefix}\n{option}"
        return hashed_string

    # Here are some methods that can be usefully overridden in subclasses.
    # It is strongly recommended not to override the methods above this line (excluding __init__)
    
    def display(self, *args, **kwargs) -> None:
        """
        Displays last message as a formatted string.
        """
        print(self.last_message_fstring)
        return

    def process_output(self, *args, **kwargs) -> None:
        """
        Full pipeline of OpenAI output processing.
        """
        msg = self.get_answer()
        self.append(msg)
        self.display(*args, **kwargs) 
        return

    def fprocess_output(self, *args, **kwargs) -> None:
        """
        Full pipeline of OpenAI output processing.
        """
        msg = self.fget_answer()
        self.append(msg)
        self.display(*args, **kwargs) 
        return
    
    def input(self, *args, **kwargs) -> Message:
        """
        User input option for console mode.
        """
        if self._runtime_mode == "app":
            return
        content = input(f"[{self.username}]: ")
        msg = self.user_message(content)
        return msg
    
    def process_input(self, *args, **kwargs) -> None:
        """
        User input processing method for both app and console modes.
        """
        if self._runtime_mode == "app":
            return
        elif self._runtime_mode == "console":
            msg = self.input(*args, **kwargs)
            self.append(msg)
            return
        else:
            raise ValueError(self._mode_error)

    def fprocess_input(self, *args, **kwargs) -> None:
        """
        An alias for 'process_input' method to run with functions.
        """
        return self.process_input(*args, **kwargs)
    
    def run(self, *args, **kwargs) -> None:
        """
        Main method. Executes dialogue with chat-bot.
        Currently only console mode is implemented.
        """
        if self.usable_functions is not None:
            return self.frun(*args, **kwargs)
        if self._runtime_mode == "app":
            return self.run_in_app_mode(*args, **kwargs)
        elif self._runtime_mode == "console":
            return self.run_in_console_mode(*args, **kwargs)
        else:
            raise ValueError(self._mode_error)
            
    def frun(self, *args, **kwargs):
        """
        Main method. Executes dialogue with chat-bot.
        Currently only console mode is implemented.
        Function calls are enabled in this method 
        (instead of 'run', in which they are disabled)
        """
        if self._runtime_mode == "app":
            return self.frun_in_app_mode(*args, **kwargs)
        elif self._runtime_mode == "console":
            return self.frun_in_console_mode(*args, **kwargs)
        else:
            raise ValueError(self._mode_error)
    
    def run_in_app_mode(self, *args, **kwargs) -> None:
        """
        Runs bot in mobile application mode.
        """
        pass
    
    def run_in_console_mode(self, *args, **kwargs) -> None:
        """
        Runs bot in console mode.
        """
        while not self.is_over:
            self.process_output(*args, **kwargs)
            self.process_input(*args, **kwargs)
            self.verify_context()
        else:
            self.end_conversation(*args, **kwargs)
        return
    
    def frun_in_app_mode(self, *args, **kwargs) -> None:
        """
        Runs bot in mobile application mode.
        """
        pass
    
    def frun_in_console_mode(self, *args, **kwargs) -> None:
        """
        Runs bot in console mode.
        """
        while not self.is_over:
            self.fprocess_output(*args, **kwargs)
            self.fprocess_input(*args, **kwargs)
            self.verify_context()
        else:
            self.end_conversation(*args, **kwargs)
        return

    def end_conversation(self, *args, **kwargs) -> None:
        content = "Был рад помочь!"
        msg = self.bot_message(content)
        self.append(msg)
        self.display(*args, **kwargs)
        self.verify_context()
        self.chat.to_disk()
        return